@@ -1,5 +1,4 @@
 import json
-
 from .decorator import ensure_table_exists, ensure_connection, handle_transactions, sql_injection_safe, enforce_resource_limits
 import pyexasol
 import dask.dataframe as dd
@@ -7,34 +6,31 @@ import pandas as pd
 import requests
 import csv
 import xlwt
-
-
-def _get_source_type(source):
-    if isinstance(source, pd.DataFrame):
-        return 'pandas'
-    elif isinstance(source, dd.DataFrame):
-        return 'dask'
-    elif isinstance(source, str):
-        if source.endswith('.csv'):
-            return 'csv'
-        elif source.endswith('.json'):
-            return 'json'
-        elif source.endswith('.xml'):
-            return 'xml'
-        elif source.endswith('.xls') or source.endswith('.xlsx'):
-            return 'excel'
-        else:
-            try:
-                json.loads(source)
-                return 'json_str'
-            except json.JSONDecodeError:
-                raise ValueError(
-                    'Unable to determine source type. Please provide a valid file path or a valid DataFrame.')
-    else:
-        raise ValueError('Unable to determine source type. Please provide a valid file path or a valid DataFrame.')
+import math
 
 
 class Features(pyexasol.ExaConnection):
+    PROTOCOL_V1 = 1
+    PROTOCOL_V2 = 2
+    PROTOCOL_V3 = 3
+
+    DEFAULT_PORT = 8563
+    DEFAULT_AUTOCOMMIT = True
+
+    DEFAULT_CONNECTION_TIMEOUT = 10
+    DEFAULT_SOCKET_TIMEOUT = 30
+    DEFAULT_QUERY_TIMEOUT = 0
+
+    DEFAULT_FETCHMANY_SIZE = 10000
+    DEFAULT_FETCH_SIZE_BYTES = 5 * 1024 * 1024
+
+    DRIVER_NAME = 'PyEXASOL'
+
+    LOGGER_FILENAME_TIMESTAMP_FORMAT = '%Y%m%d_%H%M%S_%f'
+    LOGGER_MAX_JSON_LENGTH = 20000
+
+    EXCEPTION_QUERY_TEXT_MAX_LENGTH = 20000
+
     DATATYPE_MAPPING = {
         'int8': 'TINYINT',
         'int16': 'SMALLINT',
@@ -54,108 +50,101 @@ class Features(pyexasol.ExaConnection):
         'timedelta[ns]': 'INTERVAL'
     }
 
-    def __init__(self, connection_params):
-        self.conn = None
-        self.cursor = None
-        self.connect(connection_params)
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.conn = pyexasol.connect(**connection_params)
+
+        self.options = {}
+        self.options['autocommit'] = self.DEFAULT_AUTOCOMMIT
+        self.options['snapshot_transactions'] = None
+        self.options['connection_timeout'] = self.DEFAULT_CONNECTION_TIMEOUT
+        self.options['socket_timeout'] = self.DEFAULT_SOCKET_TIMEOUT
+        self.options['query_timeout'] = self.DEFAULT_QUERY_TIMEOUT
+        self.options['compression'] = False
+        self.options['encryption'] = True
+        self.options['fetch_size_bytes'] = self.DEFAULT_FETCH_SIZE_BYTES
+        self.options['quote_ident'] = False
+        self.options['json_lib'] = 'json'
+        self.options['verbose_error'] = True
+        self.options['debug'] = False
+        self.options['debug_logdir'] = None
+        self.options['udf_output_bind_address'] = None
+        self.options['udf_output_connect_address'] = None
+        self.options['udf_output_dir'] = None
+        self.options['http_proxy'] = None
+        self.options['client_name'] = None
+        self.options['client_version'] = None
+        self.options['client_os_username'] = None
+        self.options['protocol_version'] = self.PROTOCOL_V3
+        self.options['dsn'] = None
+        self.options['user'] = None
+        self.options['password'] = None
+        self.options['schema'] = ''
+        self.options['fetch_dict'] = False
+        self.options['fetch_mapper'] = None
+        self.options['lower_ident'] = False
+        self.options['websocket_sslopt'] = None
+        self.options['access_token'] = None
+        self.options['refresh_token'] = None
 
-    @enforce_resource_limits
-    @ensure_connection
-    @sql_injection_safe
     def connect(self, connection_params):
         self.conn = pyexasol.connect(**connection_params)
-        self.cursor = self.conn.cursor()
 
-    @enforce_resource_limits
-    @ensure_connection
-    @sql_injection_safe
     def close(self):
-        self.cursor.close()
         self.conn.close()
         self.conn = None
-        self.cursor = None
 
-    @enforce_resource_limits
-    @ensure_connection
-    @ensure_table_exists
-    @handle_transactions
-    @sql_injection_safe
     def execute(self, sql, **kwargs):
-        return self.cursor.execute(sql, **kwargs)
+        return self.conn.execute(sql, **kwargs)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @sql_injection_safe
     def fetchall(self, **kwargs):
-        return self.cursor.fetchall(**kwargs)
+        return self.execute.fetchall(**kwargs)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @sql_injection_safe
     def fetchone(self, **kwargs):
-        return self.cursor.fetchone(**kwargs)
+        return self.execute.fetchone(**kwargs)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @sql_injection_safe
     def fetchmany(self, size=None, **kwargs):
-        return self.cursor.fetchmany(size=size, **kwargs)
+        return self.execute.fetchmany(size=size, **kwargs)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @sql_injection_safe
     def commit(self):
         return self.conn.commit()
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @sql_injection_safe
     def rollback(self):
         return self.conn.rollback()
 
-    def _convert_to_dask_df(self, source, source_type):
-        if source_type == 'pandas_df':
-            return dd.from_pandas(source, npartitions=self.npartitions)
-        elif source_type == 'csv':
-            return dd.read_csv(source, blocksize=self.blocksize)
-        elif source_type == 'json':
-            return dd.read_json(source, blocksize=self.blocksize)
-        elif source_type == 'xml':
-            return dd.read_xml(source, blocksize=self.blocksize)
-        elif source_type == 'excel':
-            return dd.read_excel(source, blocksize=self.blocksize)
-        else:
-            raise ValueError(f"Unrecognized source type: {source_type}")
-
-    @enforce_resource_limits
-    @ensure_connection
-    def get_column_names(self, table_name):
+    def get_column_names(self, source):
         """Get the column names for a table, query, or CTE in Exasol."""
-        self.cursor.execute(f"SELECT * FROM {table_name} LIMIT 0")
-        column_names = [d[0] for d in self.cursor.description]
-        return column_names
+        if isinstance(source, pd.DataFrame):
+            # Create the table from the DataFrame
+            return source.columns.tolist()
+        elif isinstance(source, dd.DataFrame):
+            return source.compute().columns.tolist()
+        else:
+            return self.export_to_dask(f"SELECT * FROM {source} LIMIT 0").columns.tolist()
 
-    @enforce_resource_limits
-    @ensure_connection
     def table_exists(self, table_name):
         # Check if the table exists by querying the EXA_ALL_TABLES system view
-        result = self.cursor.execute(f"SELECT 1 FROM EXA_ALL_TABLES WHERE TABLE_NAME = '{table_name}'")
+        result = self.conn.execute(f"SELECT 1 FROM EXA_ALL_TABLES WHERE TABLE_NAME = '{table_name}'")
         return bool(result.fetchone())
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
+    def get_column_data_type_from_df(self, source, column):
+        exa_data_type = self.DATATYPE_MAPPING[str(source[column].dtype)]
+        if isinstance(source, dd.DataFrame):
+            source = source.compute()
+        if exa_data_type == 'VARCHAR':
+            col_len = source[column].str.len().max()
+            potenz = 10 ** (len(str(col_len)) - 1 if len(str(col_len)) > 1 else 1)
+            col_max = math.ceil((col_len / potenz) + 0.5) * potenz
+            exa_data_type = f'VARCHAR({col_max})'
+        return exa_data_type
+
     def create_table_from_df(self, table_name, df):
+        import math
         # Create a list of column definitions
         column_defs = []
         for col, dtype in zip(df.columns, df.dtypes):
             # Get the Exasol data type for the column
-            exa_data_type = self.DATATYPE_MAPPING[str(dtype)]
+            exa_data_type = self.get_column_data_type_from_df(df, col)
             # Add the column definition to the list
             column_defs.append(f"{col} {exa_data_type}")
         # Join the column definitions with a comma separator
@@ -167,96 +156,34 @@ class Features(pyexasol.ExaConnection):
             )
         """
         # Execute the CREATE TABLE statement
-        self.cursor.execute(create_stmt)
+        self.conn.execute(create_stmt)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
     def create_table_from_table(self, target_table, source_table):
         # Build the CREATE TABLE AS SELECT statement
         create_stmt = f"CREATE TABLE {target_table} AS SELECT * FROM {source_table}"
         # Execute the CREATE TABLE AS SELECT statement
-        self.cursor.execute(create_stmt)
+        self.conn.execute(create_stmt)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
     def add_columns(self, table_name, column_names):
         # Iterate over the column names
         for column_name in column_names:
             # Build the ALTER TABLE ADD COLUMN statement
-            alter_stmt = f"ALTER TABLE {table_name} ADD COLUMN {column_name} VARCHAR"
+            alter_stmt = f"ALTER TABLE {table_name} ADD COLUMN {column_name[0]} {column_name[1]}"
             # Execute the ALTER TABLE ADD COLUMN statement
-            self.cursor.execute(alter_stmt)
+            self.conn.execute(alter_stmt)
 
-    @enforce_resource_limits
-    @ensure_connection
     def get_column_data_type(self, table_name, column_name):
         # Query the EXA_ALL_COLUMNS system view to get the data type for the column
-        result = self.cursor.execute(
-            f"SELECT COLUMN_TYPE FROM EXA_ALL_COLUMNS WHERE COLUMN_TABLE = '{table_name}' AND COLUMN_NAME = '{column_name}'")
+        result = self.conn.execute(f"SELECT COLUMN_TYPE FROM EXA_ALL_COLUMNS WHERE COLUMN_TABLE = '{table_name}' AND COLUMN_NAME = '{column_name}'")
         # Return the data type from the result
         return result.fetchone()[0]
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
     def export_to_dask(self, sql, chunksize=10000):
-        """Export data from Exasol to a Dask DataFrame using HTTP transport."""
-        # Set up HTTP connection to Exasol
-        self.conn.http_setup()
+        return dd.from_pandas(self.conn.export_to_pandas(query_or_table=sql), chunksize=chunksize)
 
-        # Send HTTP request to Exasol to execute the SQL query
-        response = requests.post(self.conn.http_url, json={'sql': sql})
-
-        # Get the column names and data types from the response
-        column_names = response.json()['column_names']
-        column_types = response.json()['column_types']
-
-        # Convert the column types to Dask data types
-        dtypes = {name: self.DATATYPE_MAPPING[typ] for name, typ in zip(column_names, column_types)}
-
-        # Create an empty Dask DataFrame with the correct column names and data types
-        df = dd.from_pandas(pd.DataFrame(columns=column_names), dtype=dtypes)
-
-        # Iterate over the data chunks in the response and append them to the Dask DataFrame
-        for chunk in response.iter_content(chunk_size=chunksize):
-            chunk_df = pd.read_json(chunk, orient='records')
-            df = df.append(dd.from_pandas(chunk_df, dtype=dtypes))
-
-        return df
-
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @ensure_table_exists
-    def import_from_dask(self, df, table_name, if_exists='fail'):
-        """Import data from a Dask DataFrame to Exasol using HTTP transport."""
-        # Set up HTTP connection to Exasol
-        self.conn.http_setup()
-
-        # Get the column names and data types from the Dask DataFrame
-        column_names = df.columns.tolist()
-        dtypes = df.dtypes.to_dict()
-
-        # Convert the Dask data types to Exasol data types
-        column_types = [self.DATATYPE_MAPPING[typ] for name, typ in dtypes.items()]
-
-        # Create a table in Exasol with the correct column names and data types
-        create_table_sql = f'CREATE TABLE {table_name} ({", ".join(f"{name} {typ}" for name, typ in zip(column_names, column_types))})'
-        self.cursor.execute(create_table_sql)
-
-        # Iterate over the partitions in the Dask DataFrame and insert the data into the Exasol table
-        for partition in df.to_delayed():
-            # Convert the partition to a Pandas DataFrame
-            partition_df = partition.compute()
-
-            # Convert the Pandas DataFrame to a list of dictionaries
-            data = [row.to_dict() for _, row in partition_df.iterrows()]
-
-            # Send an HTTP request to Exasol to insert the data into the table
-            url = self.conn.http_url + f'/{table_name}'
-            requests.post(url, json=data)
+    def import_from_dask(self, source, target_table):
+        source = source.compute()
+        self.conn.import_from_pandas(source, target_table)
 
     @staticmethod
     def to_pandas(df):
@@ -273,16 +200,6 @@ class Features(pyexasol.ExaConnection):
         """Convert a Dask DataFrame to a CSV file."""
         df.to_csv(path, index=False)
 
-    @staticmethod
-    def csv_to_excel(self, csv_file, excel_file):
-        """Convert a CSV stream to an Excel file."""
-        # Create a Pandas DataFrame from the CSV stream
-        df = pd.read_csv(csv_file)
-
-        # Use the Pandas 'to_excel' method to write the DataFrame to an Excel file
-        df.to_excel(excel_file, index=False)
-
-    @staticmethod
     def to_xml(self, df, path):
         """Convert a Dask DataFrame to an XML file."""
         # Open a file handle to the destination XML file
@@ -309,24 +226,42 @@ class Features(pyexasol.ExaConnection):
             # Write the XML footer
             xml_file.write('</root>\n')
 
-    def to_excel(self, df, path):
+    def to_excel(self, dd, path):
         """Convert a Dask DataFrame to an Excel file."""
-        # Open a file handle to the destination Excel file
-        with open(f'{path}.xlsx', 'wb') as excel_file:
-            # Use the Dask 'to_csv' method to write the DataFrame to a CSV stream
-            df.to_csv(excel_file, index=False)
+        df = dd.compute()
+        return df.to_excel(path, index=False)
 
-            # Seek back to the beginning of the stream
-            excel_file.seek(0)
+    def to_sql_select(self, source, target_table='tmp_tbl'):
+        # Prüfe, ob der source ein dask.DataFrame oder ein pandas.DataFrame ist
+        if isinstance(source, dd.DataFrame):
+            df = source.compute()
+        elif isinstance(source, pd.DataFrame):
+            df = source
 
-            # Use the 'csv_to_excel' function to convert the CSV stream to an Excel file
-            self.csv_to_excel(excel_file, excel_file)
+        # Erstelle eine Liste von Spaltennamen und ihren Datentypen
+        columns = [col.replace('-', '_').replace(' ', '_').replace('.', '_').replace('__', '_').upper() for col in df.columns.tolist()]
+        name_type = []
+        for col, name in zip(df.columns.tolist(), columns):
+            ent = {}
+            ent['df_col_name'] = col
+            ent['tbl_col_name'] = name
+            ent['is_number'] = all(isinstance(value, int) for value in df[col])
+            name_type.append(ent)
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @ensure_table_exists
-    def merge_tables(self, source, source_type, target_table, primary_columns, exclude_columns=[]):
+        # Erstelle eine Liste von SELECT-Statements für jede Zeile im DataFrame
+        selects = []
+        for row in df.values:
+            select = 'SELECT '
+            for i, col in enumerate(name_type):
+                select += "{val} AS {col}, ".format(**{"val": str(row[i]) if col['is_number'] else "'" + str(row[i]) + "'", "col": col['tbl_col_name']})
+            selects.append(select)
+
+        select_stmt = f"SELECT * FROM\n(\n"
+        select_stmt += "\nUNION ALL\n".join([select.rstrip(', ') for select in selects])
+        select_stmt += f'\n) AS {target_table}'
+        return select_stmt
+
+    def merge_tables(self, source, source_type, target_table, primary_columns=['@@@@@@@'], exclude_columns=[]):
         """Merge a table or query with an Exasol table, updating common columns and inserting new ones."""
         try:
             # Get the column names for the target table
@@ -340,16 +275,15 @@ class Features(pyexasol.ExaConnection):
             elif source_type == "query":
                 # Get the column names for the source query
                 source = f'({source})'
-                self.cursor.execute(f"SELECT * FROM {source} LIMIT 0")
-                source_columns = [d[0] for d in self.cursor.description]
+                source_columns = [d[0] for d in self.get_column_names(source)]
                 common_columns = list(set(source_columns) & set(target_columns))
 
+            # Remove the excluded columns from the list of common columns
+            all_common_columns = [col for col in common_columns if col not in exclude_columns]
+
             # Remove the primary columns from the list of common columns
             common_columns = [col for col in common_columns if col not in primary_columns]
 
-            # Remove the excluded columns from the list of common columns
-            common_columns = [col for col in common_columns if col not in exclude_columns]
-
             # Build the ON clause for the MERGE INTO statement
             on_clause = " AND ".join([f"TGT.{col} = SRC.{col}" for col in primary_columns])
 
@@ -357,7 +291,7 @@ class Features(pyexasol.ExaConnection):
             update_clause = ", ".join([f"TGT.{col} = SRC.{col}" for col in common_columns])
 
             # Build the INSERT clause for the MERGE INTO statement
-            insert_clause = ", ".join(common_columns)
+            insert_clause = ", ".join(all_common_columns)
 
             # Build the full MERGE INTO statement
             merge_stmt = f"""
@@ -372,7 +306,7 @@ class Features(pyexasol.ExaConnection):
             """
 
             # Execute the MERGE INTO statement
-            result = self.cursor.execute(merge_stmt)
+            result = self.conn.execute(merge_stmt)
         except Exception as e:
             # Print the error message and raise the exception
             print(f"Error: {e}")
@@ -381,19 +315,32 @@ class Features(pyexasol.ExaConnection):
             # Print the number of rows affected by the MERGE INTO statement
             print(f"{result.rowcount} rows affected")
 
-    @enforce_resource_limits
-    @ensure_connection
-    @handle_transactions
-    @ensure_table_exists
-    def merge_from_external(self, target_table, source, primary_columns, source_type=None):
+    def merge_from_external(self, target_table, source, primary_columns=['@@@@@@@'], chunksize=10000):
         try:
-            # Check if the source type is not provided
-            if source_type is None:
-                # Get the source type
-                source_type = _get_source_type(source)
-
             # Convert the source to a Dask DataFrame
-            source_df = self._convert_to_dask_df(source, source_type)
+            if isinstance(source, pd.DataFrame):
+                source_df = dd.from_pandas(source, chunksize=chunksize)
+            elif isinstance(source, dd.DataFrame):
+                source_df = source
+            elif isinstance(source, str):
+                if source.endswith('.csv'):
+                    source_df = dd.read_csv(source)
+                elif source.endswith('.json'):
+                    source_df = dd.read_json(source)
+                elif source.endswith('.xml'):
+                    source_df = dd.read_html(source)
+                elif source.endswith('.xls') or source.endswith('.xlsx'):
+                    source_df = dd.read_excel(source)
+                else:
+                    try:
+                        json_str = json.loads(source)
+                        source_df = dd.read_json(json_str)
+                    except json.JSONDecodeError:
+                        raise ValueError(
+                            'Unable to determine source type. Please provide a valid file path or a valid DataFrame.')
+            else:
+                raise ValueError(
+                    'Unable to determine source type. Please provide a valid file path or a valid DataFrame.')
 
             # Build the ON clause
             on_clause = " AND ".join([f"TGT.{col} = SRC.{col}" for col in primary_columns])
@@ -406,7 +353,8 @@ class Features(pyexasol.ExaConnection):
             common_columns = list(set(target_columns) & set(source_columns))
 
             # Build the UPDATE clause
-            update_clause = ", ".join([f"TGT.{col} = SRC.{col}" for col in common_columns])
+            update_clause = ", ".join(
+                [f"TGT.{col} = SRC.{col}" for col in common_columns if col not in primary_columns])
 
             # Build the INSERT clause
             insert_clause = ", ".join(common_columns)
@@ -414,7 +362,7 @@ class Features(pyexasol.ExaConnection):
             # Build the full MERGE INTO statement
             merge_stmt = f"""
                 MERGE INTO {target_table} TGT
-                USING {source_df} SRC
+                USING {self.to_sql_select(source_df, 'test')} SRC
                 ON {on_clause}
                 WHEN MATCHED THEN
                     UPDATE SET {update_clause}
@@ -424,7 +372,7 @@ class Features(pyexasol.ExaConnection):
             """
 
             # Execute the MERGE INTO statement
-            result = self.cursor.execute(merge_stmt)
+            result = self.conn.execute(merge_stmt)
         except Exception as e:
             # Print the error message and raise the exception
             print(f"Error: {e}")
@@ -432,3 +380,4 @@ class Features(pyexasol.ExaConnection):
         else:
             # Print the number of rows affected by the MERGE INTO statement
             print(f"{result.rowcount} rows affected")
+
